// Copyright (C) 2020  Wilko Manger
//
// This Source Code Form is subject to the terms of the Mozilla Public
// License, v. 2.0. If a copy of the MPL was not distributed with this
// file, You can obtain one at http://mozilla.org/MPL/2.0/.

import 'package:collection/collection.dart';

import '../../identifier.dart';
import 'ephemeral_event.dart';

import 'receipt_event.dart';
import 'typing_event.dart';

class Ephemeral implements Iterable<EphemeralEvent> {
  final Map<Type, EphemeralEvent> _map;

  Ephemeral(Iterable<EphemeralEvent> events)
      : _map = Map.fromEntries(
          events.map(
            (e) => MapEntry(e.runtimeType, e),
          ),
        );

  factory Ephemeral.fromJson(Map<String, dynamic> json, {RoomId roomId}) {
    if (json['events'] == null) {
      return Ephemeral([]);
    }

    final ephemeralEvents = json['events'] as List<dynamic>;

    return Ephemeral(
      ephemeralEvents.map(
        (e) => EphemeralEvent.fromJson(e, roomId: roomId),
      ),
    );
  }

  EphemeralEvent operator [](Type type) => _map[type];

  T get<T extends EphemeralEvent>() => this[T];

  bool containsType<T extends EphemeralEvent>() => any((e) => e is T);

  ReceiptEvent get receiptEvent => this[ReceiptEvent];

  TypingEvent get typingEvent => this[TypingEvent];

  Ephemeral copyWith({Iterable<EphemeralEvent> events}) {
    return Ephemeral(
      events ?? _map,
    );
  }

  Ephemeral merge(Ephemeral other) {
    if (other == null) return this;

    return copyWith(
      events: mergeMaps<Type, EphemeralEvent>(
        _map,
        other._map,
        value: (thisEvent, otherEvent) =>
            thisEvent is ReceiptEvent && otherEvent is ReceiptEvent
                ? thisEvent.merge(otherEvent)
                : otherEvent,
      ).values,
    );
  }

  @override
  bool any(bool Function(EphemeralEvent element) test) => _map.values.any(test);

  @override
  Iterable<R> cast<R>() => _map.values.cast<R>();

  @override
  bool contains(Object element) => _map.values.contains(element);

  @override
  EphemeralEvent elementAt(int index) => _map.values.elementAt(index);

  @override
  bool every(
    bool Function(EphemeralEvent element) test,
  ) =>
      _map.values.every(test);

  @override
  Iterable<T> expand<T>(Iterable<T> Function(EphemeralEvent element) f) =>
      _map.values.expand(f);

  @override
  EphemeralEvent get first => _map.values.first;

  @override
  EphemeralEvent firstWhere(
    bool Function(EphemeralEvent element) test, {
    EphemeralEvent Function() orElse,
  }) =>
      _map.values.firstWhere(test, orElse: orElse);

  @override
  T fold<T>(
    T initialValue,
    T Function(T previousValue, EphemeralEvent element) combine,
  ) =>
      _map.values.fold(initialValue, combine);

  @override
  Iterable<EphemeralEvent> followedBy(Iterable<EphemeralEvent> other) =>
      _map.values.followedBy(other);

  @override
  void forEach(void Function(EphemeralEvent element) f) =>
      _map.values.forEach(f);

  @override
  bool get isEmpty => _map.values.isEmpty;

  @override
  bool get isNotEmpty => _map.values.isNotEmpty;

  @override
  Iterator<EphemeralEvent> get iterator => _map.values.iterator;

  @override
  String join([String separator = '']) => _map.values.join(separator);

  @override
  EphemeralEvent get last => _map.values.last;

  @override
  EphemeralEvent lastWhere(
    bool Function(EphemeralEvent element) test, {
    EphemeralEvent Function() orElse,
  }) =>
      _map.values.lastWhere(test, orElse: orElse);

  @override
  int get length => _map.values.length;

  @override
  Iterable<T> map<T>(T Function(EphemeralEvent e) f) => _map.values.map(f);

  @override
  EphemeralEvent reduce(
          EphemeralEvent Function(
    EphemeralEvent value,
    EphemeralEvent element,
  )
              combine) =>
      _map.values.reduce(combine);

  @override
  EphemeralEvent get single => _map.values.single;

  @override
  EphemeralEvent singleWhere(
    bool Function(EphemeralEvent element) test, {
    EphemeralEvent Function() orElse,
  }) =>
      _map.values.singleWhere(test, orElse: orElse);

  @override
  Iterable<EphemeralEvent> skip(int count) => _map.values.skip(count);

  @override
  Iterable<EphemeralEvent> skipWhile(
    bool Function(EphemeralEvent value) test,
  ) =>
      _map.values.skipWhile(test);

  @override
  Iterable<EphemeralEvent> take(int count) => _map.values.take(count);

  @override
  Iterable<EphemeralEvent> takeWhile(
    bool Function(EphemeralEvent value) test,
  ) =>
      _map.values.takeWhile(test);

  @override
  List<EphemeralEvent> toList({bool growable = true}) =>
      _map.values.toList(growable: growable);

  @override
  Set<EphemeralEvent> toSet() => _map.values.toSet();

  @override
  Iterable<EphemeralEvent> where(bool Function(EphemeralEvent element) test) =>
      _map.values.where(test);

  @override
  Iterable<T> whereType<T>() => _map.values.whereType<T>();
}
