// Copyright (C) 2020  Wilko Manger
// Copyright (C) 2019  Mathieu Velten
//
// This Source Code Form is subject to the terms of the Mozilla Public
// License, v. 2.0. If a copy of the MPL was not distributed with this
// file, You can obtain one at http://mozilla.org/MPL/2.0/.

import 'dart:async';

import 'package:meta/meta.dart';

import 'homeserver.dart';
import 'identifier.dart';
import 'store/store.dart';

/// [Identifiable] objects that can possibly be loaded.
///
/// This is only what's currently loaded in memory. If you want to load more
/// (either from a [Store] or remotely, use [load].
@immutable
abstract class Loadables<I extends Id, S extends Identifiable<I>>
    implements Iterable<S> {
  final Context context;

  final List<S> _values;

  Loadables(
    Iterable<S> iterable, {
    this.context,
  }) : _values = List.of(iterable);

  Loadables.empty({this.context}) : _values = [];

  Future<void> load({int count = 20}) async {}

  @override
  bool any(bool Function(S element) test) => _values.any(test);

  @override
  Iterable<R> cast<R>() => _values.cast<R>();

  @override
  bool contains(Object element) => _values.contains(element);

  @override
  S elementAt(int index) => _values.elementAt(index);

  @override
  bool every(bool Function(S element) test) => _values.every(test);

  @override
  Iterable<T> expand<T>(Iterable<T> Function(S element) f) => _values.expand(f);

  @override
  S get first => _values.first;

  @override
  S firstWhere(bool Function(S element) test, {S Function() orElse}) =>
      _values.firstWhere(test, orElse: orElse);

  @override
  T fold<T>(T initialValue, T Function(T previousValue, S element) combine) =>
      _values.fold(initialValue, combine);

  @override
  Iterable<S> followedBy(Iterable<S> other) => _values.followedBy(other);

  @override
  void forEach(void Function(S element) f) => _values.forEach(f);

  @override
  bool get isEmpty => _values.isEmpty;

  @override
  bool get isNotEmpty => _values.isNotEmpty;

  @override
  Iterator<S> get iterator => _values.iterator;

  @override
  String join([String separator = '']) => _values.join(separator);

  @override
  S get last => _values.last;

  @override
  S lastWhere(bool Function(S element) test, {S Function() orElse}) =>
      _values.lastWhere(test, orElse: orElse);

  @override
  int get length => _values.length;

  @override
  Iterable<T> map<T>(T Function(S e) f) => _values.map(f);

  @override
  S reduce(S Function(S value, S element) combine) => _values.reduce(combine);

  @override
  S get single => _values.single;

  @override
  S singleWhere(bool Function(S element) test, {S Function() orElse}) =>
      _values.singleWhere(test, orElse: orElse);

  @override
  Iterable<S> skip(int count) => _values.skip(count);

  @override
  Iterable<S> skipWhile(bool Function(S value) test) => _values.skipWhile(test);

  @override
  Iterable<S> take(int count) => _values.take(count);

  @override
  Iterable<S> takeWhile(bool Function(S value) test) => _values.takeWhile(test);

  @override
  List<S> toList({bool growable = true}) => _values.toList(growable: growable);

  @override
  Set<S> toSet() => _values.toSet();

  @override
  Iterable<S> where(bool Function(S element) test) => _values.where(test);

  @override
  Iterable<T> whereType<T>() => _values.whereType<T>();
}
