// Copyright (C) 2020  Wilko Manger
//
// This Source Code Form is subject to the terms of the Mozilla Public
// License, v. 2.0. If a copy of the MPL was not distributed with this
// file, You can obtain one at http://mozilla.org/MPL/2.0/.

import 'package:meta/meta.dart';

import 'room/rooms.dart';
import 'device.dart';
import 'homeserver.dart';
import 'identifier.dart';
import 'notifications/pushers.dart';
import 'store/store.dart';
import 'updater.dart';
import 'user.dart';

/// A user which is authenticated and can send messages, join rooms etc.
@immutable
class MyUser extends User {
  final Context context;

  @override
  final UserId id;

  @override
  final String name;

  @override
  final Uri avatarUrl;

  final String accessToken;

  final String syncToken;

  final Device currentDevice;

  final Rooms rooms;

  final Pushers pushers;

  final bool isLoggedOut;

  MyUser({
    this.context,
    this.id,
    this.name,
    this.avatarUrl,
    this.accessToken,
    this.syncToken,
    this.currentDevice,
    this.rooms,
    this.isLoggedOut,
  }) : pushers = context != null && id != null
            ? Pushers(Context(myId: id, homeserver: context.homeserver))
            : null;

  static Future<MyUser> fromStore(Store store) async {
    final opened = await store.open();
    if (!opened) {
      return null;
    }

    return await store.getMyUser();
  }

  MyUser.base({
    @required Homeserver homeserver,
    @required UserId id,
    @required String accessToken,
  }) : this(
          context: Context(
            homeserver: homeserver,
            myId: id,
          ),
          id: id,
          name: null,
          avatarUrl: null,
          accessToken: accessToken,
          currentDevice: null,
          rooms: Rooms.empty(
            context: Context(
              homeserver: homeserver,
              myId: id,
            ),
          ),
          isLoggedOut: false,
        );

  MyUser copyWith({
    UserId id,
    String name,
    Uri avatarUrl,
    String accessToken,
    String syncToken,
    Device currentDevice,
    Rooms rooms,
    bool isLoggedOut,
  }) {
    return MyUser(
      context: context,
      id: id ?? this.id,
      name: name ?? this.name,
      avatarUrl: avatarUrl ?? this.avatarUrl,
      accessToken: accessToken ?? this.accessToken,
      syncToken: syncToken ?? this.syncToken,
      currentDevice: currentDevice ?? this.currentDevice,
      rooms: rooms ?? this.rooms,
      isLoggedOut: isLoggedOut ?? this.isLoggedOut,
    );
  }

  MyUser merge(MyUser other) {
    if (other == null) return this;

    return copyWith(
      id: other.id,
      name: other.name,
      avatarUrl: other.avatarUrl,
      accessToken: other.accessToken,
      syncToken: other.syncToken,
      currentDevice:
          currentDevice?.merge(other.currentDevice) ?? other.currentDevice,
      rooms: rooms?.merge(other.rooms) ?? other.rooms,
      isLoggedOut: other.isLoggedOut,
    );
  }

  @override
  List<Object> get props => [
        context,
        id,
        name,
        avatarUrl,
        accessToken,
        currentDevice,
        rooms,
        isLoggedOut,
      ];

  /// Get all invites for this user. Note that for now this will load
  /// all rooms to memory.
  /*Future<List<Invite>> get invites async =>
      (await _rooms.get(where: (r) => r is InvitedRoom))
          .map((r) => Invite._(scope, r))
          .toList(growable: false);*/

  /// Set the display name of this user to the given value.
  ///
  /// Returns the [Update] where [MyUser] has it's name set to [name]`,
  /// if successful.
  Future<RequestUpdate<MyUser>> setName(String name) async =>
      context.updater.setName(name: name);

  /// Invalidates the access token of the user. Makes all
  /// [MyUser] calls unusable.
  ///
  /// Returns the [Update] where [MyUser] has logged out, if successful.
  Future<RequestUpdate<MyUser>> logout() => context.updater.logout();

  /// Send all unsent messages still in the [Store].
  /*Future<void> sendAllUnsent() async {
    for (Room room in await rooms.get()) {
      if (room is JoinedRoom) {
        for (RoomEvent event in await _store.getUnsentEvents(room)) {
          await for (final _ in room.send(
            event.content,
            transactionId: event.transactionId,
          )) {}
        }
      }
    }
  }*/

  Future<Uri> upload({
    @required Stream<List<int>> bytes,
    @required int length,
    @required String contentType,
    String fileName,
  }) =>
      context.homeserver.upload(
        as: id,
        bytes: bytes,
        length: length,
        contentType: contentType,
        fileName: fileName,
      );

  Stream<Update> get updates => context.updater.updates;

  bool get isSyncing => context.updater.syncer.isSyncing;

  void startSync({Duration maxRetryAfter}) =>
      context.updater.syncer.start(maxRetryAfter: maxRetryAfter);

  Future<void> stopSync() => context.updater.syncer.stop();
}
