// Copyright (C) 2019  Wilko Manger
// Copyright (C) 2019  Mathieu Velten
//
// This Source Code Form is subject to the terms of the Mozilla Public
// License, v. 2.0. If a copy of the MPL was not distributed with this
// file, You can obtain one at http://mozilla.org/MPL/2.0/.

import 'package:meta/meta.dart';

import '../event/room/room_event.dart';
import '../identifier.dart';
import '../room/member/member.dart';
import '../room/room.dart';
import '../homeserver.dart';
import '../my_user.dart';

/// Stores all data (rooms, users, events) somewhere.
abstract class Store {
  bool get isOpen;

  /// Returns true if opened successfully.
  Future<bool> open();
  Future<void> close();

  /// Gets the currently stored [MyUser].
  ///
  /// If [Homeserver] was not provided on, it will create a new instance and
  /// associate it to the returned [MyUser] via its [Context]
  ///
  /// Returns null if there's no user stored.
  Future<MyUser> getMyUser();

  /// Save [MyUser] and all it's data completely.
  Future<void> setMyUserDelta(MyUser myUser);

  /// Returned sorted alphabetically based on the room ID.
  ///
  /// See [Rooms].
  Future<Iterable<Room>> getRooms();

  Future<Room> getRoom(RoomId id);

  /// Returned sorted based on [RoomEvent.time], newest first.
  ///
  /// Will also return a list of MemberChangeEvents needed to render the
  /// returned timeline events.
  ///
  /// If [fromTime] is provided, will skip all events with a time greater
  /// than [fromTime].
  Future<Messages> getMessages(
    RoomId roomId, {
    int count = 20,
    DateTime fromTime,
  });

  Future<Iterable<RoomEvent>> getUnsentEvents();

  /// Returned sorted based on since.
  ///
  /// See [Members].
  Future<Iterable<Member>> getMembers(
    RoomId roomId, {
    int count = 20,
    DateTime fromTime,
  });
}

/// Combination of RoomEvents and MemberChangeEvents needed to render
/// them. Named after the API.
@immutable
class Messages {
  final Iterable<RoomEvent> events;
  final Iterable<Member> state;

  Messages(this.events, this.state);
}
