// Copyright (C) 2020  Wilko Manger
//
// This Source Code Form is subject to the terms of the Mozilla Public
// License, v. 2.0. If a copy of the MPL was not distributed with this
// file, You can obtain one at http://mozilla.org/MPL/2.0/.

import 'dart:async';
import 'dart:io';

import 'package:mime/mime.dart';
import 'package:pedantic/pedantic.dart';
import 'package:async/async.dart';
import 'package:meta/meta.dart';
import 'package:image/image.dart';
import 'package:synchronized/synchronized.dart';

import 'event/ephemeral/ephemeral.dart';
import 'event/ephemeral/receipt_event.dart';
import 'event/ephemeral/typing_event.dart';
import 'event/event.dart';
import 'event/room/message_event.dart';
import 'event/room/redaction_event.dart';
import 'event/room/room_event.dart';
import 'event/room/state/room_creation_event.dart';
import 'event/room/state/state_event.dart';
import 'homeserver.dart';
import 'identifier.dart';
import 'room/member/member_timeline.dart';
import 'room/member/membership.dart';
import 'my_user.dart';
import 'room/room.dart';
import 'room/rooms.dart';
import 'store/store.dart';
import 'room/timeline.dart';
import 'util/random.dart';

/// Manages updates to [MyUser].
class Updater {
  final Homeserver homeserver;
  final Store store;

  /// Most up to date instance of our user.
  MyUser user;

  _Syncer _syncer;
  _Syncer get syncer => _syncer ??= _Syncer(this);

  final _subject = StreamController<Update>.broadcast();
  Stream<Update> get updates => _subject.stream;

  Updater(this.user, this.homeserver, this.store);

  final _lock = Lock();

  /// Send out an update, with a new user which is the current [user]
  /// merged with [delta].
  Future<Update> _update<U extends Update>(
    MyUser delta,
    U Function(MyUser user, MyUser delta) createUpdate,
  ) async {
    return await _lock.synchronized(() {
      user = user.merge(delta);

      unawaited(store.setMyUserDelta(delta.copyWith(id: user.id)));

      final update = createUpdate(user, delta);
      _subject.add(update);
      return update;
    });
  }

  Future<RequestUpdate<MyUser>> setName({@required String name}) async {
    await homeserver.api.profile.putDisplayName(
      accessToken: user.accessToken,
      userId: user.id.toString(),
      value: name,
    );

    return await _update(
      MyUser(name: name),
      (user, delta) => RequestUpdate(
        user,
        delta,
        data: user,
        deltaData: delta,
      ),
    );
  }

  Future<RequestUpdate<MemberTimeline>> kick(
    UserId id, {
    @required RoomId from,
  }) async {
    if (user.rooms[from].members[id]?.membership == Membership.kicked) {
      return RequestUpdate.fromUpdate(
        await updates.first,
        data: (u) => u.rooms[from].memberTimeline,
        deltaData: (u) => u?.rooms[from]?.memberTimeline,
      );
    }

    await homeserver.api.rooms.kick(
      accessToken: user.accessToken,
      roomId: from.toString(),
      userId: id.toString(),
    );

    return RequestUpdate.fromUpdate(
      await updates.firstWhere(
        (u) => u.delta.rooms[from].members.current.kicked.any(
          (m) => m.id == id,
        ),
      ),
      data: (u) => u.rooms[from].memberTimeline,
      deltaData: (u) => u.rooms[from].memberTimeline,
    );
  }

  Stream<RequestUpdate<Timeline>> send(
    RoomId roomId,
    EventContent content, {
    String transactionId,
    String stateKey = '',
  }) async* {
    transactionId ??= randomString();

    final eventArgs = RoomEventArgs(
      id: EventId(transactionId),
      roomId: roomId,
      time: DateTime.now(),
      senderId: user.id,
      sentState: SentState.unsent,
      transactionId: transactionId,
    );

    var event = RoomEvent.fromContent(content, eventArgs);

    if (event == null || event is RoomCreationEvent) {
      // TODO: Allow any custom event to be sent
      throw ArgumentError('This event type cannot be send.');
    }

    yield await _update(
      MyUser(
        rooms: Rooms([
          Room(
            id: roomId,
            timeline: Timeline([event]),
          ),
        ]),
      ),
      (user, delta) => RequestUpdate(
        user,
        delta,
        data: user.rooms[roomId].timeline,
        deltaData: user.rooms[roomId].timeline,
      ),
    );

    // TODO: Support for web
    // Upload images from image message events that have a file uri
    if (event is ImageMessageEvent && event.content.url.scheme == 'file') {
      final imageEvent = event as ImageMessageEvent;

      final file = File(
        imageEvent.content.url.toFilePath(windows: Platform.isWindows),
      );

      final fileName = file.path.split(Platform.pathSeparator).last;

      final matrixUrl = await user.upload(
        bytes: file.openRead(),
        length: await file.length(),
        contentType: lookupMimeType(file.path),
        fileName: fileName,
      );

      final image = decodeImage(file.readAsBytesSync());

      // TODO: Add copyWith
      event = RoomEvent.fromContent(
        ImageMessage(
          url: matrixUrl,
          body: imageEvent.content.body,
          inReplyToId: imageEvent.content.inReplyToId,
          info: ImageInfo(
            width: image.width,
            height: image.height,
          ),
        ),
        eventArgs,
      );
    }

    Map<String, dynamic> body;
    if (event is StateEvent) {
      body = await homeserver.api.rooms.sendState(
        accessToken: user.accessToken,
        roomId: roomId.toString(),
        eventType: event.type,
        stateKey: stateKey,
        content: event.content.toJson(),
      );
    } else {
      body = await homeserver.api.rooms.send(
        accessToken: user.accessToken,
        roomId: roomId.toString(),
        eventType: event.type,
        transactionId: transactionId,
        content: event.content.toJson(),
      );
    }

    final eventId = EventId(body['event_id']);

    final sentEvent = RoomEvent.fromContent(
      content,
      eventArgs.copyWith(
        id: eventId,
        sentState: SentState.sent,
      ),
    );

    yield await _update(
      MyUser(
        rooms: Rooms([
          Room(
            id: roomId,
            timeline: Timeline([sentEvent]),
          ),
        ]),
      ),
      (user, delta) => RequestUpdate(
        user,
        delta,
        data: user.rooms[roomId].timeline,
        deltaData: user.rooms[roomId].timeline,
      ),
    );
  }

  Future<RequestUpdate<List<Receipt>>> markRead({
    @required RoomId roomId,
    @required EventId until,
    bool receipt = true,
  }) async {
    if (receipt) {
      final isReadAlready = user.rooms[roomId].readReceipts.any(
        (receipt) => receipt.eventId == until,
      );

      if (isReadAlready) {
        return RequestUpdate.fromUpdate(
          await updates.first,
          data: (u) => u.rooms[roomId].readReceipts,
          deltaData: (u) => u.rooms[roomId]?.readReceipts ?? [],
        );
      }
    }

    await homeserver.api.rooms.readMarkers(
      accessToken: user.accessToken,
      roomId: roomId.toString(),
      fullyRead: until.toString(),
      read: receipt ? until.toString() : null,
    );

    final relevantUpdate = receipt
        ? await updates.firstWhere(
            (update) =>
                update.delta.rooms[roomId]?.readReceipts?.any(
                  (receipt) => receipt.eventId == until,
                ) ??
                false,
          )
        : await updates.first;

    return RequestUpdate.fromUpdate(
      relevantUpdate,
      data: (u) => u.rooms[roomId].readReceipts,
      deltaData: (u) => u.rooms[roomId].readReceipts,
    );
  }

  Future<RequestUpdate<MyUser>> logout() async {
    return await _lock.synchronized(() async {
      await homeserver.api.logout();
      await syncer.stop();

      return _update(
        MyUser(isLoggedOut: true),
        (user, delta) => RequestUpdate(
          user,
          delta,
          data: user,
          deltaData: delta,
        ),
      );
    });
  }

  Future<RequestUpdate<Rooms>> loadRooms({int count}) async {
    final rooms = await store.getRooms();

    return await _update(
      MyUser(rooms: Rooms(rooms)),
      (user, delta) => RequestUpdate(
        user,
        delta,
        data: user.rooms,
        deltaData: delta.rooms,
      ),
    );
  }

  Future<RequestUpdate<Timeline>> loadEvents({
    @required RoomId roomId,
    int count = 20,
  }) async {
    final currentRoom = user.rooms[roomId];

    final messages = await store.getMessages(
      roomId,
      count: count,
      fromTime: currentRoom.timeline.last.time,
    );

    var timeline = Timeline(messages.events);

    var memberTimeline = MemberTimeline(messages.state);

    if (timeline.length < count) {
      count -= timeline.length;

      final body = await homeserver.api.rooms.messages(
        accessToken: user.accessToken,
        roomId: roomId.toString(),
        limit: count,
        from: currentRoom.timeline.previousBatch,
        filter: {
          'lazy_load_members': true,
        },
      );

      timeline = timeline.merge(
        Timeline.fromJson(
          (body['chunk'] as List<dynamic>).cast(),
          roomId: roomId,
          previousBatch: body['end'],
          previousBatchSetBySync: false,
        ),
      );

      if (body.containsKey('state')) {
        memberTimeline = memberTimeline.merge(
          MemberTimeline.fromEvents([
            ...timeline,
            ...(body['state'] as List<dynamic>)
                .cast<Map<String, dynamic>>()
                .map((e) => RoomEvent.fromJson(e, roomId: roomId)),
          ]),
        );
      }
    }

    final newRoom = Room(
      id: currentRoom.id,
      timeline: timeline,
      memberTimeline: memberTimeline,
    );

    return await _update(
      MyUser(rooms: Rooms([newRoom])),
      (user, delta) => RequestUpdate(
        user,
        delta,
        data: user.rooms[newRoom.id].timeline,
        deltaData: delta.rooms[newRoom.id].timeline,
      ),
    );
  }

  Future<RequestUpdate<MemberTimeline>> loadMembers({
    @required RoomId roomId,
    int count = 10,
  }) async {
    final currentRoom = user.rooms[roomId];

    final members = await store.getMembers(
      roomId,
      fromTime: currentRoom.memberTimeline.last.since,
      count: count,
    );

    var memberTimeline = MemberTimeline(members);

    if (members.length < count) {
      count -= members.length;

      final body = await homeserver.api.rooms.members(
        accessToken: user.accessToken,
        roomId: roomId.toString(),
        at: currentRoom.timeline.previousBatch,
      );

      final events = (body['chunk'] as List<dynamic>)
          .cast<Map<String, dynamic>>()
          .map((e) => RoomEvent.fromJson(e, roomId: roomId));

      memberTimeline = memberTimeline.merge(
        MemberTimeline.fromEvents(events),
      );
    }

    final newRoom = Room(
      id: roomId,
      memberTimeline: memberTimeline,
    );

    return await _update(
      MyUser(rooms: Rooms([newRoom])),
      (user, delta) => RequestUpdate(
        user,
        delta,
        data: user.rooms[newRoom.id].memberTimeline,
        deltaData: user.rooms[newRoom.id].memberTimeline,
      ),
    );
  }

  Future<RequestUpdate<Ephemeral>> setIsTyping({
    @required RoomId roomId,
    @required bool isTyping,
    Duration timeout = const Duration(seconds: 30),
  }) async {
    await homeserver.api.rooms.typing(
      accessToken: user.accessToken,
      roomId: roomId.toString(),
      userId: user.id.toString(),
      typing: isTyping,
      timeout: timeout.inMilliseconds,
    );

    return RequestUpdate.fromUpdate(
      await user.updates.firstWhere((u) {
        final containsMe = u.delta.rooms[roomId]?.ephemeral
            ?.get<TypingEvent>()
            ?.content
            ?.typerIds
            ?.contains(user.id);

        return containsMe == null ? false : isTyping ? containsMe : !containsMe;
      }),
      data: (u) => u.rooms[roomId].ephemeral,
      deltaData: (u) => u.rooms[roomId]?.ephemeral,
    );
  }

  /// Note: Will return RequestUpdate<Pushers> in the future.
  Future<void> setPusher(Map<String, dynamic> pusher) async {
    await homeserver.api.pushers.set(
      accessToken: user.accessToken,
      body: pusher,
    );
  }

  Future<void> _processSync(Map<String, dynamic> body) async {
    final roomDeltas = await _processRooms(body);

    await _update(
      MyUser(syncToken: body['next_batch'], rooms: Rooms(roomDeltas)),
      (user, delta) => SyncUpdate(user, delta),
    );
  }

  /// Returns list of room delta.
  Future<List<Room>> _processRooms(Map<String, dynamic> body) async {
    final jRooms = body['rooms'];

    const join = 'join';
    const invite = 'invite';
    const leave = 'leave';

    Future<List<Room>> process(
      Map<String, dynamic> rooms, {
      @required String type,
    }) async {
      final roomDeltas = <Room>[];

      if (rooms != null) {
        for (final entry in rooms.entries) {
          final roomId = RoomId(entry.key);
          final json = entry.value;

          var currentRoom = user.rooms[roomId] ?? await store.getRoom(roomId);

          var isNewRoom = false;
          if (currentRoom == null) {
            isNewRoom = true;
            currentRoom = Room.base(
              context: Context(
                myId: user.id,
                homeserver: homeserver,
              ),
              myMembership: Membership.parse(type),
              id: roomId,
            );
          }

          var roomDelta = Room.fromJson(json, id: roomId);

          // Set previous batch to null if it wasn't set by sync before
          if (!(currentRoom.timeline.previousBatchSetBySync ?? true)) {
            roomDelta = roomDelta.copyWith(
              // We can't use copyWith because we're setting previousBatch to
              // null again
              timeline: Timeline(
                roomDelta.timeline,
                context: null,
                previousBatch: null,
                previousBatchSetBySync: false,
              ),
            );
          }

          final accountData = body['account_data'];
          // Process account data
          if (accountData != null) {
            final events = accountData['events'] as List<dynamic>;

            final event = events.firstWhere(
              (event) => event['type'] == 'm.direct',
              orElse: () => null,
            );

            if (event != null) {
              final content = event['content'] as Map<String, dynamic>;

              for (final entry in content.entries) {
                final userId = entry.key;
                final roomIds = entry.value;

                if (UserId.isValidFullyQualified(userId) &&
                    roomIds.contains(roomId.toString())) {
                  roomDelta = roomDelta.copyWith(directUserId: UserId(userId));
                  break;
                }
              }
            }
          }

          // Process redactions
          // TODO: Redaction deltas
          for (final event
              in currentRoom.timeline.whereType<RedactionEvent>()) {
            final redactedId = event.redacts;

            final original = currentRoom.timeline[redactedId];
            if (original != null && original is! RedactedEvent) {
              final newTimeline = currentRoom.timeline.merge(
                Timeline([
                  ...currentRoom.timeline.where((e) => e.id != redactedId),
                  RedactedEvent.fromRedaction(
                      redaction: event, original: original),
                ]),
              );

              roomDelta = roomDelta.copyWith(timeline: newTimeline);
            }
          }

          if (isNewRoom) {
            roomDelta = currentRoom.merge(roomDelta);
          }

          roomDeltas.add(roomDelta);
        }

        return roomDeltas;
      }

      return null;
    }

    return [
      ...await process(jRooms[join], type: join),
      ...await process(jRooms[invite], type: invite),
      ...await process(jRooms[leave], type: leave),
    ];
  }
}

/// An update to [MyUser], either because of a sync or because of a request.
@immutable
abstract class Update {
  final MyUser user;

  /// The delta [MyUser]. All properties of [delta] are null except
  /// those that are changed with this update.
  ///
  /// This is useful to find out what changed between an update.
  final MyUser delta;

  Update._(this.user, this.delta);
}

/// An update caused by a sync.
class SyncUpdate extends Update {
  SyncUpdate(
    MyUser user,
    MyUser delta,
  ) : super._(user, delta);
}

extension UpdatesExtension on Stream<Update> {
  Future<Update> get firstSync => firstWhere((u) => u is SyncUpdate);

  Stream<Update> get onlySync => where((u) => u is SyncUpdate);
}

/// An update caused by a request, which has the relevant updated [data] at
/// hand for easy access, and also a [deltaData].
class RequestUpdate<T> extends Update {
  final T data;
  final T deltaData;

  RequestUpdate(
    MyUser user,
    MyUser deltaUser, {
    @required this.data,
    @required this.deltaData,
  }) : super._(user, deltaUser);

  RequestUpdate.fromUpdate(
    Update update, {
    @required T Function(MyUser user) data,
    @required T Function(MyUser delta) deltaData,
  }) : this(
          update.user,
          update.delta,
          data: data(update.user),
          deltaData: deltaData(update.delta),
        );
}

class _Syncer {
  final Updater updater;

  Homeserver get homeserver => updater.homeserver;
  Store get store => updater.store;
  MyUser get user => updater.user;

  bool _isSyncing = false;

  bool get isSyncing => _isSyncing;

  _Syncer(this.updater);

  Future<void> _syncFuture;
  CancelableOperation<Map<String, dynamic>> _cancelableSyncOnceResponse;

  /// Syncs data with the user's [homeserver]. Use the [sync] [Observable] to
  /// listen to sync changes.
  void start({
    Duration maxRetryAfter = const Duration(seconds: 30),
  }) {
    if (user.isLoggedOut) {
      throw StateError('The user can not be logged out');
    }

    _syncFuture = _startSync(maxRetryAfter: maxRetryAfter);
  }

  bool _shouldStopSync = false;

  Future<void> _startSync({
    Duration maxRetryAfter,
  }) async {
    maxRetryAfter ??= const Duration(seconds: 30);

    _shouldStopSync = false;
    _isSyncing = true;

    // This var is used to implements exponential backoff
    // until it reaches maxRetryAfter
    var retryAfter = 500;

    while (!_shouldStopSync) {
      final body = await _sync(
        timeout: Duration(seconds: 10),
      );

      if (_shouldStopSync) return;

      if (body == null) {
        await Future.delayed(Duration(milliseconds: retryAfter));

        if (_shouldStopSync) return;

        retryAfter = (retryAfter * 1.5).floor();
        if (retryAfter > maxRetryAfter.inMilliseconds) {
          retryAfter = maxRetryAfter.inMilliseconds;
        }
      } else {
        await updater._processSync(body);

        // Reset exponential backoff.
        retryAfter = 500;
      }
    }
  }

  Future<Map<String, dynamic>> _sync({
    timeout = Duration.zero,
    bool fullState = false,
  }) async {
    if (user.isLoggedOut) {
      throw StateError('The user can not be logged out');
    }

    if (_shouldStopSync) return null;

    try {
      final cancelable = CancelableOperation.fromFuture(
        homeserver.api.sync(
          accessToken: user.accessToken,
          since: user.syncToken,
          fullState: fullState,
          filter: {
            'room': {
              'state': {
                'lazy_load_members': true,
              },
              'timeline': {
                'limit': 30,
              },
            },
          },
          timeout: timeout.inMilliseconds,
        ),
      );

      _cancelableSyncOnceResponse = cancelable;

      final body = await cancelable.valueOrCancellation();

      // We're cancelled
      if (body == null) return null;

      if (_shouldStopSync) return null;

      return body;
    } on Exception {
      // TODO: Handle exceptions better
      return null;
    }
  }

  Future<void> stop() async {
    _shouldStopSync = true;
    await _cancelableSyncOnceResponse?.cancel();
    await _syncFuture;
    _isSyncing = false;
  }
}
